// Copyright 2006 Nemanja Trifunovic
// Modified by Darrell Wright 2018
/*
Permission is hereby granted, free of charge, to any person or organization
obtaining a copy of the software and accompanying documentation covered by
this license (the "Software") to use, reproduce, display, distribute,
execute, and transmit the Software, and to prepare derivative works of the
Software, and to permit third-parties to whom the Software is furnished to
do so, all subject to the following:

The copyright notices in the Software and this entire statement, including
the above license grant, this restriction and the following disclaimer,
must be included in all copies of the Software, in whole or in part, and
all derivative works of the Software, unless such copies or derivative
works are solely in the form of machine-executable object code generated by
a source language processor.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

#pragma once

#include "core.h"

#include <daw/cpp_17.h>
#include <daw/daw_traits.h>

#include <iterator>

namespace daw::utf8::unchecked {
	template<typename octet_iterator>
	constexpr octet_iterator append( uint32_t cp,
	                                 octet_iterator result ) noexcept {

		if( cp < 0x80 ) { // one octet
			*( result++ ) = static_cast<char>( cp );
		} else if( cp < 0x800 ) { // two octets
			*( result++ ) = static_cast<char>( ( cp >> 6 ) | 0xc0 );
			*( result++ ) = static_cast<char>( ( cp & 0x3f ) | 0x80 );
		} else if( cp < 0x10000 ) { // three octets
			*( result++ ) = static_cast<char>( ( cp >> 12 ) | 0xe0 );
			*( result++ ) = static_cast<char>( ( ( cp >> 6 ) & 0x3f ) | 0x80 );
			*( result++ ) = static_cast<char>( ( cp & 0x3f ) | 0x80 );
		} else { // four octets
			*( result++ ) = static_cast<char>( ( cp >> 18 ) | 0xf0 );
			*( result++ ) = static_cast<char>( ( ( cp >> 12 ) & 0x3f ) | 0x80 );
			*( result++ ) = static_cast<char>( ( ( cp >> 6 ) & 0x3f ) | 0x80 );
			*( result++ ) = static_cast<char>( ( cp & 0x3f ) | 0x80 );
		}
		return result;
	}

	template<typename octet_iterator>
	constexpr uint32_t next( octet_iterator &it ) noexcept {

		uint32_t cp = utf8::internal::mask8( *it );
		auto const length = utf8::internal::sequence_length( it );

		switch( length ) {
		case 1:
			break;
		case 2:
			++it;
			cp = ( ( cp << 6 ) & 0x7ff ) + ( ( *it ) & 0x3f );
			break;
		case 3:
			++it;
			cp = ( ( cp << 12 ) & 0xffff ) +
			     ( ( utf8::internal::mask8( *it ) << 6 ) & 0xfff );
			++it;
			cp += ( *it ) & 0x3f;
			break;
		case 4:
			++it;
			cp = ( ( cp << 18 ) & 0x1fffff ) +
			     ( ( utf8::internal::mask8( *it ) << 12 ) & 0x3ffff );
			++it;
			cp += ( utf8::internal::mask8( *it ) << 6 ) & 0xfff;
			++it;
			cp += ( *it ) & 0x3f;
			break;
		}
		++it;
		return cp;
	}

	template<typename octet_iterator>
	constexpr uint32_t peek_next( octet_iterator it ) noexcept {
		return utf8::unchecked::next( it );
	}

	template<typename octet_iterator>
	constexpr uint32_t prior( octet_iterator &it ) noexcept {
		while( utf8::internal::is_trail( *( --it ) ) ) {}

		auto temp = it;
		return utf8::unchecked::next( temp );
	}

	// Deprecated in versions that include prior, but only for the sake of
	// consistency (see utf8::previous)
	template<typename octet_iterator>
	constexpr uint32_t previous( octet_iterator &it ) noexcept {
		return utf8::unchecked::prior( it );
	}

	template<typename octet_iterator, typename distance_type>
	constexpr void advance( octet_iterator &it, distance_type n ) noexcept {
		for( distance_type i = 0; i < n; ++i ) {
			utf8::unchecked::next( it );
		}
	}

	template<typename octet_iterator>
	constexpr auto distance( octet_iterator first,
	                         octet_iterator last ) noexcept ->
	  typename std::iterator_traits<octet_iterator>::difference_type {

		typename std::iterator_traits<octet_iterator>::difference_type dist = 0;
		for( ; first < last; ++dist ) {
			utf8::unchecked::next( first );
		}
		return dist;
	}

	template<typename u16bit_iterator, typename octet_iterator>
	constexpr octet_iterator utf16to8( u16bit_iterator start, u16bit_iterator end,
	                                   octet_iterator result ) noexcept {
		while( start != end ) {
			uint32_t cp = utf8::internal::mask16( *start++ );
			// Take care of surrogate pairs first
			if( utf8::internal::is_lead_surrogate( cp ) ) {
				uint32_t trail_surrogate = utf8::internal::mask16( *start++ );
				cp = ( cp << 10 ) + trail_surrogate + internal::SURROGATE_OFFSET;
			}
			result = utf8::unchecked::append( cp, result );
		}
		return result;
	}

	template<typename u16bit_iterator, typename octet_iterator>
	constexpr u16bit_iterator utf8to16( octet_iterator start, octet_iterator end,
	                                    u16bit_iterator result ) noexcept {
		while( start < end ) {
			auto cp = utf8::unchecked::next( start );
			if( cp > 0xFFFF ) { // make a surrogate pair
				*result++ =
				  static_cast<uint16_t>( ( cp >> 10 ) + internal::LEAD_OFFSET );

				*result++ = static_cast<uint16_t>( ( cp & 0x3ff ) +
				                                   internal::TRAIL_SURROGATE_MIN );
			} else {
				*result++ = static_cast<uint16_t>( cp );
			}
		}
		return result;
	}

	template<typename octet_iterator, typename u32bit_iterator>
	constexpr octet_iterator utf32to8( u32bit_iterator start, u32bit_iterator end,
	                                   octet_iterator result ) noexcept {
		while( start != end ) {
			result = utf8::unchecked::append( *( start++ ), result );
		}

		return result;
	}

	template<typename octet_iterator, typename u32bit_iterator>
	constexpr u32bit_iterator utf8to32( octet_iterator start, octet_iterator end,
	                                    u32bit_iterator result ) noexcept {
		while( start < end ) {
			( *result++ ) = utf8::unchecked::next( start );
		}

		return result;
	}

	// The iterator class
	template<typename octet_iterator>
	class iterator {
		octet_iterator it;

	public:
		using iterator_category = std::bidirectional_iterator_tag;
		using value_type = uint32_t;
		using difference_type = std::ptrdiff_t;
		using pointer = value_type *;
		using reference = value_type &;
		using const_reference = value_type const &;

		constexpr iterator( ) noexcept(
		  std::is_nothrow_default_constructible_v<octet_iterator> ) {}

		constexpr explicit iterator( const octet_iterator &octet_it ) noexcept(
		  std::is_nothrow_copy_constructible_v<octet_iterator> )
		  : it( octet_it ) {}

		// the default "big three" are OK
		constexpr octet_iterator base( ) const
		  noexcept( std::is_nothrow_copy_constructible_v<octet_iterator> ) {
			return it;
		}

		constexpr uint32_t operator*( ) const noexcept {
			octet_iterator temp = it;
			return utf8::unchecked::next( temp );
		}

		constexpr bool operator==( const iterator &rhs ) const noexcept {
			return ( it == rhs.it );
		}

		constexpr bool operator!=( const iterator &rhs ) const noexcept {
			return !( operator==( rhs ) );
		}

		constexpr iterator &operator++( ) noexcept {
			std::advance( it, utf8::internal::sequence_length( it ) );
			return *this;
		}

		constexpr iterator operator++( int ) noexcept {
			iterator temp = *this;
			std::advance( it, utf8::internal::sequence_length( it ) );
			return temp;
		}

		constexpr iterator &operator--( ) noexcept {
			utf8::unchecked::prior( it );
			return *this;
		}

		constexpr iterator operator--( int ) noexcept {
			iterator temp = *this;
			utf8::unchecked::prior( it );
			return temp;
		}

		constexpr iterator &operator+=( difference_type n ) noexcept {
			while( n-- > 0 ) {
				++( *this );
			}
			return *this;
		}

		constexpr iterator &operator-=( difference_type n ) noexcept {
			while( n-- > 0 ) {
				--( *this );
			}
			return *this;
		}

	}; // class iterator

	template<typename octet_iterator>
	iterator( octet_iterator ) -> iterator<octet_iterator>;

	template<typename octet_iterator>
	constexpr iterator<octet_iterator>
	operator+( iterator<octet_iterator> it,
	           typename iterator<octet_iterator>::difference_type n ) noexcept {
		return it += n;
	}

	template<typename octet_iterator>
	constexpr iterator<octet_iterator>
	operator-( iterator<octet_iterator> it,
	           typename iterator<octet_iterator>::difference_type n ) noexcept {
		return it -= n;
	}
} // namespace daw::utf8::unchecked
